from typing import Literal

from .utils.cst import ConflictResolution, ConflictResolver, PythonContext

resolution_mapping = {
    "e": ConflictResolution.KEEP_PREVIOUS,
    "n": ConflictResolution.KEEP_NEW,
    "m": ConflictResolution.KEEP_MERGED,
    "r": ConflictResolution.REMOVE,
    "t": ConflictResolution.NEW_AS_TODO,
}


class InteractiveResolver(ConflictResolver):
    def __init__(self, debug: bool):
        self.debug = debug

    def resolve_conflict(
        self, context: PythonContext, existing_node, new_node, merged_node
    ) -> ConflictResolution:
        if not self.debug:
            if existing_node and not new_node:
                return ConflictResolution.KEEP_PREVIOUS
            if new_node:
                return ConflictResolution.KEEP_NEW
        if existing_node:
            if new_node:
                print("Conflict detected, please decide")
                print("--------------------------------")
                print("Existing node:")
                print(context.to_source_code(existing_node).strip())
                print("New node:")
                print(context.to_source_code(new_node).strip())
                print("Will be merged to:")
                print(context.to_source_code(merged_node).strip())
                print("--------------------------------")

                print("Your decision: ")
                print("   m or enter to keep the merged value")
                print("   e  to keep the existing value")
                print("   n  to keep the new value")
                print(
                    "   t  to keep the previous value and add the new value as TODO comment"
                )
                while True:
                    inp = input().strip()
                    if inp in resolution_mapping:
                        return resolution_mapping[inp]
                    elif inp == "":
                        return ConflictResolution.KEEP_MERGED
                    print("Unknown decision")
            else:
                print("Removal detected, please decide")
                print("--------------------------------")
                print("Existing node:")
                print(context.to_source_code(existing_node).strip())
                print("--------------------------------")

                print("Your decision: ")
                print("   e or enter to keep the existing value")
                print("   r  to remove the existing value")
                while True:
                    inp = input().strip()
                    if inp in resolution_mapping:
                        return resolution_mapping[inp]
                    elif inp == "":
                        return ConflictResolution.KEEP_PREVIOUS
                    print("Unknown decision")
        else:
            if new_node:
                print("Addition detected, please decide")
                print("--------------------------------")
                print("New node:")
                print(context.to_source_code(new_node).strip())
                print("--------------------------------")

                print("Your decision: ")
                print("   n or enter to keep the new value")
                print(
                    "   t  to keep the previous value and add the new value as TODO comment"
                )
                while True:
                    inp = input().strip()
                    if inp in resolution_mapping:
                        return resolution_mapping[inp]
                    elif inp == "":
                        return ConflictResolution.KEEP_NEW
                    print("Unknown decision")
        return ConflictResolution.KEEP_MERGED


class AutomaticResolver(ConflictResolver):
    def __init__(self, resolution_type: Literal["replace", "keep", "comment"]):
        self.resolution_type = resolution_type

    def resolve_conflict(
        self, context, existing_node, new_node, merged_node
    ) -> ConflictResolution:
        if existing_node:
            if new_node:
                if self.resolution_type == "replace":
                    return ConflictResolution.KEEP_NEW
                elif self.resolution_type == "keep":
                    return ConflictResolution.KEEP_PREVIOUS
                elif self.resolution_type == "comment":
                    return ConflictResolution.NEW_AS_TODO
            else:
                return ConflictResolution.KEEP_PREVIOUS
        else:
            if new_node:
                return ConflictResolution.KEEP_NEW
        return ConflictResolution.KEEP_MERGED



from oarepo_model_builder.outputs import OutputBase


class TextOutput(OutputBase):
    TYPE = "text"

    def begin(self):
        try:
            with self.builder.filesystem.open(self.path) as f:
                self.text = self.original_data = f.read()
        except FileNotFoundError:
            self.original_data = None
            self.text = ""

    @property
    def created(self):
        return self.original_data is None

    def finish(self):
        if self.text != self.original_data:
            with self.builder.filesystem.open(self.path, "w") as f:
                f.write(self.text)



from .model import ModelOpenSearchSchema, ModelSchema
from .model_defaults import ModelDefaults
from .model_plugins import PluginConfigSchema, PluginsSchema
from .object import ObjectDynamicSchema
from .properties import ArrayItemsSchema, ObjectFieldSchema, PropertiesSchema
from .property import ObjectProperty, Property
from .property_facets import PropertyFacets
from .property_jsonschema import PropertyJSONSchema
from .property_mapping import PropertyMapping
from .property_marshmallow import (
    ModelMarshmallowSchema,
    ObjectPropertyMarshmallowSchema,
    PropertyMarshmallowSchema,
)
from .property_sample_data import ModelSampleConfiguration, PropertySampleData
from .property_sortable import PropertySortable
from .root import RootSchema
from .settings import SettingsOpenSearchSchema, SettingsPythonSchema, SettingsSchema
from .ui import ObjectPropertyUISchema, ObjectUIPropertyValidator, UIPropertyValidator

#
# Validators is a dictionary of "points" in model schema mapped to a single ma.Schema class
# or a list of ma.Schema classes. During validation, a new marshmallow class is constructed
# with all those classes as the bases for the class.
#
# To create your own extension point, see for example root.py: RootSchema
#
validators = {
    # Root of the model schema - to extend it with new
    # top-level options, add your schema here
    "root": RootSchema,
    #
    # /settings: To add your own settings key, register it here
    #
    "settings": SettingsSchema,
    #
    # /settings/opensearch: opensearch settings, currently only "version"
    "settings-opensearch": SettingsOpenSearchSchema,
    #
    # /settings/python: use-black, use-isort, ....
    # To add your own schema-wide python options, put
    # them in ma.Schema and register them here
    "settings-python": SettingsPythonSchema,
    #
    # /model: represents the content of the top-level 'model' element.
    # to add model-wide settings, register your schema here
    #
    "model": [
        ModelSchema,
        ModelDefaults,
        ModelSampleConfiguration,
        ModelMarshmallowSchema,
    ],
    #
    # /model/opensearch: defines opensearch index settings (analyzers, etc).
    # You mostly do not want to extend this
    "model-opensearch": ModelOpenSearchSchema,
    #
    # /model/ui extensibility point
    #
    "model-ui": [UIPropertyValidator, ObjectUIPropertyValidator],
    #
    # /model/marshmallow: defines marshmallow on the model-level
    "property-marshmallow-model": ModelMarshmallowSchema.ObjectOnlyMarshmallowProps,
    #
    # /model/properties, /model/properties/<element_name>/properties, ...
    # defines content of the properties (any key mapping to object-field).
    # You mostly do not want to modify this.
    "properties": PropertiesSchema,
    #
    # represents a wrapper around one item in properties.
    # You do not want to modify this
    "object-field": ObjectFieldSchema,
    #
    # represents a wrapper around an array "items" field.
    # You do not want to modify this
    "array-items": ArrayItemsSchema,
    #
    # This is the base schema for properties/aaa or array item.
    # Register schema here to add type-independent extensions
    "property": [Property],
    #
    # <property>/ui extensibility point
    #
    "property-ui": [UIPropertyValidator],
    #
    # <property>[/type=object]/ui extensibility point
    #
    "object-property-ui": [ObjectUIPropertyValidator],
    #
    # This is the base schema for properties/aaa
    # Register schema here to add type-independent extensions that are used only on object members
    "object-property": [ObjectProperty],
    #
    # This is the base schema for ...aaa[type=array]/items
    # Register schema here to add type-independent extensions that are used only on array items
    "array-item": [],
    #
    # <items>/ui extensibility point
    #
    "array-item-ui": [],
    #
    # An extension point for properties/aaa/facets
    "property-facets": PropertyFacets,
    #
    # An extension point for properties/aaa/sample
    "property-sample": PropertySampleData,
    #
    # An extension point for properties/aaa/jsonschema
    "property-jsonschema": PropertyJSONSchema,
    #
    # An extension point for properties/aaa/mapping
    "property-mapping": PropertyMapping,
    #
    # An extension point for properties/aaa/marshmallow
    "property-marshmallow": PropertyMarshmallowSchema,
    #
    # An extension point for properties/aaa/sortable
    "property-sortable": PropertySortable,
    #
    # property-by-type-xxx defines an extension point for schemas
    # that are dependent on property type. The schema is taken automatically
    # from the datatype (DataType.ModelSchema) property, so primarily define
    # your configuration in there.
    #
    # However, if you need to separate concerns and your extension works on
    # just a single data type, add it to this group of extensions
    #
    # This defines extra marshmallow fields for object/nested containers (schema-class, generate, ...)
    "property-by-type-object": [
        ObjectPropertyMarshmallowSchema,
        ObjectPropertyUISchema,
        ObjectDynamicSchema,
    ],
    "property-by-type-nested": [
        ObjectPropertyMarshmallowSchema,
        ObjectPropertyUISchema,
        ObjectDynamicSchema,
    ],
    #
    # An extension point for properties/aaa/marshmallow
    "property-marshmallow-object": ObjectPropertyMarshmallowSchema.ObjectMarshmallowProps,
    #
    # /model/plugins schema - defines model builder plugins that
    # are available. Extend this if you add a new "type" of plugins.
    "plugins-schema": PluginsSchema,
    #
    # defines switching on/off plugins of a single "type", for example /model/plugins/builder/enabled
    "plugin-schema": PluginConfigSchema,
}



from ..outputs.python import PythonOutput
from .python import PythonBuilder


class SetupPyBuilder(PythonBuilder):
    TYPE = "setup_py"

    def finish(self):
        super().finish()

        python_output: PythonOutput = self.builder.get_output("python", "setup.py")
        python_output.merge(
            "setup_py", {"settings": self.settings, "current_model": self.current_model}
        )



from marshmallow import fields

from ..utils.facet_helpers import facet_definition, facet_name
from .datatypes import DataType


class NumberDataType(DataType):
    def marshmallow_validators(self):
        validators = []
        ranges = {}
        for param, schema in (
            ("min", "minimumExclusive"),
            ("max", "maximumExclusive"),
            ("min_inclusive", "minimum"),
            ("max_inclusive", "maximum"),
        ):
            if schema in self.definition:
                ranges[param] = self.definition[schema]

        if ranges:
            params = ", ".join(f"{k}={v}" for k, v in ranges.items())
            validators.append(f"ma_validate.Range({params})")

        return validators

    def get_facet(self, stack, parent_path):
        key, field = facet_definition(self)
        path = parent_path
        if len(parent_path) > 0 and self.key:
            path = parent_path + "." + self.key
        elif self.key:
            path = self.key
        if field:
            return [{"facet": field, "path": facet_name(path)}]
        else:
            return [{"facet": f'TermsFacet(field="{path}")', "path": facet_name(path)}]


class IntegerDataType(NumberDataType):
    marshmallow_field = "ma_fields.Integer"
    schema_type = "integer"
    model_type = "integer"

    class ModelSchema(DataType.ModelSchema):
        minimum = fields.Integer(required=False)
        exclusiveMinimum = fields.Integer(required=False)
        maximum = fields.Integer(required=False)
        exclusiveMaximum = fields.Integer(required=False)
        enum = fields.List(fields.Integer(), required=False)


class FloatDataType(NumberDataType):
    marshmallow_field = "ma_fields.Float"
    schema_type = "number"
    model_type = "float"

    class ModelSchema(DataType.ModelSchema):
        minimum = fields.Float(required=False)
        exclusiveMinimum = fields.Float(required=False)
        maximum = fields.Float(required=False)
        exclusiveMaximum = fields.Float(required=False)
        enum = fields.List(fields.Float(), required=False)


class DoubleDataType(NumberDataType):
    marshmallow_field = "ma_fields.Float"
    schema_type = "number"
    model_type = "double"

    class ModelSchema(DataType.ModelSchema):
        minimum = fields.Float(required=False)
        exclusiveMinimum = fields.Float(required=False)
        maximum = fields.Float(required=False)
        exclusiveMaximum = fields.Float(required=False)
        enum = fields.List(fields.Float(), required=False)


class BooleanDataType(DataType):
    marshmallow_field = "ma_fields.Boolean"
    model_type = "boolean"

    def get_facet(self, stack, parent_path):
        key, field = facet_definition(self)
        path = parent_path
        if len(parent_path) > 0 and self.key:
            path = parent_path + "." + self.key
        elif self.key:
            path = self.key
        if field:
            return [{"facet": field, "path": facet_name(path)}]
        else:
            return [{"facet": f'TermsFacet(field="{path}")', "path": facet_name(path)}]



from oarepo_model_builder.builder import ModelBuilder
from oarepo_model_builder.outputs.jsonschema import JSONSchemaOutput
from oarepo_model_builder.outputs.mapping import MappingOutput
from oarepo_model_builder.schema import ModelSchema


def test_output_disabled():
    builder = ModelBuilder()
    schema = ModelSchema("", {"model": {"plugins": {"output": {"disable": "__all__"}}}})
    builder.set_schema(schema)
    assert builder._filter_classes([JSONSchemaOutput, MappingOutput], "output") == []


def test_output_disabled_single():
    builder = ModelBuilder()
    schema = ModelSchema(
        "", {"model": {"plugins": {"output": {"disable": ["jsonschema"]}}}}
    )
    builder.set_schema(schema)
    assert set(
        x.TYPE
        for x in builder._filter_classes([JSONSchemaOutput, MappingOutput], "output")
    ) == {"mapping"}


def test_output_enabled():
    builder = ModelBuilder()
    schema = ModelSchema(
        "",
        {
            "model": {
                "plugins": {"output": {"disable": "__all__", "enable": ["mapping"]}}
            }
        },
    )
    builder.set_schema(schema)
    assert set(
        x.TYPE
        for x in builder._filter_classes([JSONSchemaOutput, MappingOutput], "output")
    ) == {"mapping"}


def test_output_enabled_import():
    builder = ModelBuilder()
    schema = ModelSchema(
        "",
        {
            "model": {
                "plugins": {
                    "output": {
                        "include": [
                            "oarepo_model_builder.outputs.mapping:MappingOutput"
                        ]
                    }
                }
            }
        },
    )
    builder.set_schema(schema)
    assert set(x.TYPE for x in builder._filter_classes([], "output")) == {"mapping"}



import marshmallow as ma
from marshmallow import fields


class AdditionalPropertiesSchema(ma.Schema):
    type = fields.String(required=True)


class PropertyNamesSchema(ma.Schema):
    pattern = fields.String(required=True)


class ObjectDynamicSchema(ma.Schema):
    additionalProperties = fields.Nested(AdditionalPropertiesSchema, required=False)
    propertyNames = fields.Nested(PropertyNamesSchema, required=False)



from collections import namedtuple
from dataclasses import dataclass, field
from enum import Enum
from logging import getLogger
from typing import List, Set, Union

from libcst import Arg, CSTNode, Element, Module

from .mergers import module_mergers

logger = getLogger("oarepo_model_builder.cst")


class OperationPerformed(Enum):
    ADD = 1
    CHANGE = 2
    REMOVAL = 3


class ConflictResolution(Enum):
    KEEP_PREVIOUS = 1
    KEEP_NEW = 2
    KEEP_MERGED = 3
    REMOVE = 4
    NEW_AS_TODO = 5


class ConflictResolver:
    def resolve_conflict(
        self, context, existing_node, new_node, merged_node
    ) -> ConflictResolution:
        return ConflictResolution.KEEP_PREVIOUS


@dataclass
class PythonContextItem:
    existing_node: CSTNode
    new_node: CSTNode
    operations: Set[OperationPerformed] = field(default_factory=set, init=False)
    new_as_comment: bool = field(default=False, init=False)


@dataclass
class PythonContext:
    cst: Module
    conflict_resolver: ConflictResolver = None
    stack: List[PythonContextItem] = field(default_factory=list, init=False)

    REMOVED = object()

    def to_source_code(self, node):
        if not node:
            return ""
        return self.cst.code_for_node(node)

    def push(self, existing_node: CSTNode, new_node: CSTNode):
        self.stack.append(
            PythonContextItem(existing_node=existing_node, new_node=new_node)
        )

    def pop(self):
        self.stack.pop()

    @property
    def top(self):
        return self.stack[-1]

    def decide(
        self,
        existing_node: Union[CSTNode, None],
        new_node: Union[CSTNode, None],
        merged_node: Union[CSTNode, None],
    ) -> Union[CSTNode, object]:
        logger.debug(
            "\nDecide called with operations %s on node %s %s",
            self.top.operations,
            type(existing_node or new_node).__name__,
            id(existing_node),
        )
        logger.debug("Existing: %s", self.to_source_code(existing_node))
        logger.debug("New     : %s", self.to_source_code(new_node))
        logger.debug("Merged  : %s", self.to_source_code(merged_node))

        top = self.top
        decider_called = False
        decision = None
        if existing_node is self.top.existing_node:
            # processing existing node. If there was any decision in children, do not decide here
            if top.operations:
                decision = ConflictResolution.KEEP_MERGED
            if len(self.stack) > 1:
                top = self.stack[-2]
        if _get_source_code(self, existing_node) == _get_source_code(self, new_node):
            decision = ConflictResolution.KEEP_PREVIOUS
        if decision is None:
            if existing_node:
                decision = ConflictResolution.KEEP_PREVIOUS
            else:
                decision = ConflictResolution.KEEP_NEW

            if self.conflict_resolver:
                decision = self.conflict_resolver.resolve_conflict(
                    self, existing_node, new_node, merged_node
                )
            decider_called = True

        logger.debug(
            "---> %s %s",
            "<decider returned>" if decider_called else "<automatic>",
            decision,
        )
        logger.debug("")
        if decision == ConflictResolution.KEEP_PREVIOUS:
            return existing_node
        elif decision == ConflictResolution.KEEP_NEW:
            top.operations.add(OperationPerformed.ADD)
            return new_node
        elif decision == ConflictResolution.KEEP_MERGED:
            top.operations.add(OperationPerformed.CHANGE)
            return merged_node
        elif decision == ConflictResolution.REMOVE:
            top.operations.add(OperationPerformed.REMOVAL)
            return self.REMOVED
        elif decision == ConflictResolution.NEW_AS_TODO:
            top.new_as_comment = True
            return existing_node
        raise RuntimeError("Unknown decision")


node_with_type = namedtuple("node_with_type", "node, type")


class MergerBase:
    def merge(self, context: PythonContext, existing_node, new_node):
        try:
            context.push(existing_node, new_node)
            if existing_node and new_node and isinstance(new_node, type(existing_node)):
                ret = self.merge_internal(context, existing_node, new_node)
            else:
                ret = existing_node or new_node
            return context.decide(existing_node, new_node, ret)
        finally:
            context.pop()

    def identity(self, context, node):
        print(
            f"Warning: using naive merger for {type(node)}"
            f">>>\n{context.to_source_code(node).strip()}\n<<<"
        )
        return node

    def merge_internal(self, context: PythonContext, existing_node, new_node):
        raise NotImplementedError()


class IdentityBaseMerger(MergerBase):
    def merge_internal(self, context: PythonContext, existing_node, new_node):
        return existing_node

    def identity(self, context, node):
        return node


class IdentityMerger(IdentityBaseMerger):
    def merge_internal(self, context: PythonContext, existing_node, new_node):
        return existing_node

    def identity(self, context, node):
        print(
            f"Warning: using naive merger for {type(node)}"
            f">>>\n{context.to_source_code(node).strip()}\n<<<"
        )
        return node


def merge(context: PythonContext, existing_node, new_node, mergers=None):
    if not mergers:
        mergers = module_mergers

    merger = mergers.get(type(existing_node or new_node), IdentityMerger())
    return merger.merge(context, existing_node, new_node)


def _get_source_code(context, node):
    if isinstance(node, (Arg, Element)):
        node = node.value
    return context.to_source_code(node)



import json
import logging
from hashlib import sha256

from marshmallow import fields

from oarepo_model_builder.utils.jinja import split_package_base_name, split_package_name
from oarepo_model_builder.utils.python_name import convert_name_to_python_class
from oarepo_model_builder.validation import InvalidModelException, model_validator

from .datatypes import DataType

log = logging.getLogger("datatypes")


class ObjectDataType(DataType):
    schema_type = "object"
    mapping_type = "object"
    marshmallow_field = "ma_fields.Nested"
    ui_marshmallow_field = "ma_fields.Nested"
    model_type = "object"

    class ModelSchema(DataType.ModelSchema):
        properties = fields.Nested(
            lambda: model_validator.validator_class("properties", strict=False)()
        )

    def prepare(self, context):
        super().prepare(context)
        self._prepare_schema_class(
            self.definition.setdefault("marshmallow", {}),
            split_package_name(self.model.record_schema_class),
            context,
        )
        ui = self.definition.setdefault("ui", {})
        self._prepare_schema_class(
            ui.setdefault("marshmallow", {}),
            split_package_name(self.model.record_ui_schema_class),
            context,
            suffix="UISchema",
        )

    def _prepare_schema_class(
        self,
        marshmallow_definition,
        package_name,
        context,
        suffix="Schema",
    ):
        if 'schema-class' in marshmallow_definition and marshmallow_definition['schema-class'] is None:
            return
        schema_class = marshmallow_definition.get("schema-class", None)
        if schema_class:
            absolute_class_name = self._get_class_name(package_name, schema_class)
            marshmallow_definition["schema-class"] = absolute_class_name
            return

        if not schema_class:
            if self.stack.top.schema_element_type == "items":
                schema_class_base = self.stack[-2].key + "Item"
            else:
                schema_class_base = self.key
            schema_class = convert_name_to_python_class(schema_class_base) + suffix

        schema_class = self._get_class_name(package_name, schema_class)

        fingerprint = sha256(
            json.dumps(
                self.definition, sort_keys=True, default=lambda x: repr(x)
            ).encode("utf-8")
        ).hexdigest()

        # separate ui and normal marshmallow classes
        class_cache = context.setdefault(f"marshmallow-class-cache-{suffix}", {})

        schema_class = self._find_unique_schema_class(
            class_cache, schema_class, fingerprint
        )
        log.debug(
            "%s: fp %s, schema class %s", self.stack.path, fingerprint, schema_class
        )

        class_cache[schema_class] = fingerprint

        marshmallow_definition["schema-class"] = schema_class

        return marshmallow_definition

    def _find_unique_schema_class(self, known_classes, schema_class, fingerprint):
        parent = self.stack[:-1]
        package_name, class_name = split_package_base_name(schema_class)
        orig_schema_class = schema_class

        # generate unique class name (if duplicates are found) by using more and more from the path
        while True:
            if schema_class not in known_classes:
                # first occurrence, just return
                return schema_class
            if known_classes[schema_class] == fingerprint:
                # same name and fingerprint, just return
                return schema_class
            if not parent:
                # could not resolve parent
                break
            top = parent[-1]
            parent = parent[:-1]
            # if top is not property, can't add to name, so continue with its parent
            if top.schema_element_type != "property":
                continue
            class_name = convert_name_to_python_class(top.key) + class_name
            schema_class = f"{package_name}.{class_name}"

        # generate unique class name (if duplicates are found) by appending a number
        package_name, class_name = split_package_base_name(orig_schema_class)
        for i in range(1, 100):
            schema_class = f"{package_name}.{class_name}{i}"
            if schema_class not in known_classes:
                # first occurrence, just return
                return schema_class
            if known_classes[schema_class] == fingerprint:
                # same name and fingerprint, just return
                return schema_class

        raise InvalidModelException(
            f"Too many marshmallow classes with name {schema_class}. Please specify your own class names"
        )

    def _get_class_name(self, package_name: str, class_name: str):
        if "." not in class_name:
            return f"{package_name}.{class_name}"
        if class_name.startswith("."):
            package_path = package_name.split(".")
            while class_name.startswith("."):
                if package_path:
                    package_path = package_path[:-1]
                class_name = class_name[1:]
            if package_path:
                class_name = f"{'.'.join(package_path)}.{class_name}"
        return class_name

    def get_facet(self, stack, parent_path):
        if not stack:
            return None
        path = parent_path
        if len(parent_path) > 0 and self.key:
            path = parent_path + "." + self.key
        elif self.key:
            path = self.key
        return stack[0].get_facet(stack[1:], path)


class NestedDataType(ObjectDataType):
    schema_type = "object"
    mapping_type = "nested"
    marshmallow_field = "ma_fields.Nested"
    ui_marshmallow_field = "ma_fields.Nested"
    model_type = "nested"

    def get_facet(self, stack, parent_path):
        if not stack:
            return None
        path = parent_path
        if len(parent_path) > 0 and self.key:
            path = parent_path + "." + self.key
        elif self.key:
            path = self.key
        facet_obj = stack[0].get_facet(stack[1:], path)

        nested_arr = []
        for f in facet_obj:
            nested_arr.append(
                {
                    "facet": f'NestedLabeledFacet(path ="{path}", nested_facet = {f["facet"]})',
                    "path": f["path"],
                }
            )
        return nested_arr


class FlattenDataType(DataType):
    schema_type = "object"
    mapping_type = "object"
    marshmallow_field = "ma_fields.Raw"
    ui_marshmallow_field = "ma_fields.Raw"
    model_type = "flattened"

    def get_facet(self, stack, parent_path):
        pass

    def prepare(self, context):
        # not indexing for now as
        mapping = self.definition.setdefault("mapping", {})
        mapping.setdefault("enabled", False)
        super().prepare(context)


class ArrayDataType(DataType):
    schema_type = "array"
    mapping_type = None
    marshmallow_field = "ma_fields.List"
    ui_marshmallow_field = "ma_fields.List"
    model_type = "array"

    class ModelSchema(DataType.ModelSchema):
        items = fields.Nested(
            lambda: model_validator.validator_class("array-items", strict=False)()
        )
        uniqueItems = fields.Boolean(required=False)
        minItems = fields.Integer(required=False)
        maxItems = fields.Integer(required=False)

    def get_facet(self, stack, parent_path):
        if not stack:
            return None
        path = parent_path
        if len(parent_path) > 0 and self.key:
            path = parent_path + "." + self.key
        elif self.key:
            path = self.key
        return stack[0].get_facet(stack[1:], path)



from oarepo_model_builder.model_preprocessors import ModelPreprocessor
from oarepo_model_builder.utils.camelcase import snake_case
from oarepo_model_builder.utils.deepmerge import deepmerge

MARSHMALLOW_SCHEMA_BASE_CLASS = "ma.Schema"


class InvenioModelPreprocessor(ModelPreprocessor):
    TYPE = "invenio"

    def transform(self, schema, settings):
        model = schema.current_model
        deepmerge(settings, {"python": {"templates": {}}})
        deepmerge(
            model,
            {
                # just make sure that the templates is always there
                "marshmallow": {},
            },
        )

        record_prefix = model.record_prefix

        extension_suffix = snake_case(record_prefix)
        extension_suffix_upper = extension_suffix.upper()
        self.set(model, "extension-suffix", lambda: extension_suffix)
        self.set(model, "profile-package", lambda: "records")

        # self.set(model, "extension-suffix", lambda: f"_{record_prefix}")

        self.set(
            model,
            "record-prefix-snake",
            lambda: snake_case(model.record_prefix),
        )

        # level-1 packages

        self.set(
            model,
            "record-resources-package",
            lambda: f"{model.package}.resources.{model.profile_package}",
        )

        self.set(
            model,
            "record-services-package",
            lambda: f"{model.package}.services.{model.profile_package}",
        )

        self.set(
            model,
            "record-records-package",
            lambda: f"{model.package}.{model.profile_package}",
        )

        # config
        self.set(model, "config-package", lambda: f"{model.package}.config")
        self.set(
            model,
            "config-dummy-class",
            lambda: f"{model.package}.config.DummyClass",
        )
        # todo "config prefix"
        self.set(
            model,
            "config-resource-config-key",
            lambda: f"{model.package_base_upper}_RESOURCE_CONFIG_{extension_suffix_upper}",
        )
        self.set(
            model,
            "config-resource-class-key",
            lambda: f"{model.package_base_upper}_RESOURCE_CLASS_{extension_suffix_upper}",
        )
        self.set(
            model,
            "config-service-config-key",
            lambda: f"{model.package_base_upper}_SERVICE_CONFIG_{extension_suffix_upper}",
        )
        self.set(
            model,
            "config-service-class-key",
            lambda: f"{model.package_base_upper}_SERVICE_CLASS_{extension_suffix_upper}",
        )
        self.set(
            model,
            "config-resource-register-blueprint-key",
            lambda: f"{model.package_base_upper}_REGISTER_BLUEPRINT",
        )

        # ext
        self.set(
            model,
            "ext-class",
            lambda: f"{model.package}.ext.{record_prefix}Ext",
        )
        self.set(model, "flask-extension-name", lambda: extension_suffix)

        # cli
        self.set(
            model,
            "cli-function",
            lambda: f"{model.package}.cli.group",
        )

        # proxies
        self.set(
            model,
            "proxies-current-resource",
            lambda: f"{model.package}.proxies.current_resource",
        )
        self.set(
            model,
            "proxies-current-service",
            lambda: f"{model.package}.proxies.current_service",
        )

        # record
        self.set(
            model,
            "record-class",
            lambda: f"{model.record_records_package}.api.{record_prefix}Record",
        )
        self.set(
            model,
            "record-metadata-class",
            lambda: f"{model.record_records_package}.models.{record_prefix}Metadata",
        )
        self.set(
            model,
            "record-metadata-table-name",
            lambda: f"{record_prefix.lower()}_metadata",
        )
        #   - setup.cfg
        self.set(
            model,
            "record-mapping-setup-cfg",
            lambda: f"{extension_suffix}",
        )
        self.set(
            model,
            "record-jsonschemas-setup-cfg",
            lambda: f"{extension_suffix}",
        )
        self.set(
            model,
            "record-api-blueprints-setup-cfg",
            lambda: f"{extension_suffix}",
        )
        self.set(
            model,
            "record-blueprints-setup-cfg",
            lambda: f"{extension_suffix}",
        )

        # resource
        self.set(
            model,
            "record-resource-config-class",
            lambda: f"{model.record_resources_package}.config.{record_prefix}ResourceConfig",
        )
        self.set(
            model,
            "record-resource-class",
            lambda: f"{model.record_resources_package}.resource.{record_prefix}Resource",
        )
        self.set(
            model,
            "record-permissions-class",
            lambda: f"{model.record_services_package}.permissions.{record_prefix}PermissionPolicy",
        )

        # service
        self.set(
            model,
            "record-service-class",
            lambda: f"{model.record_services_package}.service.{record_prefix}Service",
        )
        self.set(
            model,
            "record-service-config-class",
            lambda: f"{model.record_services_package}.config.{record_prefix}ServiceConfig",
        )

        model.setdefault("record-service-config-generate-links", True)
        #   - schema
        self.set(
            model,
            "record-schema-class",
            lambda: f"{model.record_services_package}.schema.{record_prefix}Schema",
        )
        self.set(
            model,
            "record-ui-schema-class",
            lambda: f"{model.record_services_package}.ui_schema.{record_prefix}UISchema",
        )
        self.set(
            model,
            "record-ui-serializer-class",
            lambda: f"{model.record_resources_package}.ui.{record_prefix}UIJSONSerializer",
        )
        self.set(
            model,
            "record-schema-metadata-class",
            lambda: f"{model.record_services_package}.schema.{record_prefix}MetadataSchema",
        )
        self.set(
            model,
            "record-ui-schema-metadata-class",
            lambda: f"{model.record_services_package}.ui_schema.{record_prefix}MetadataUISchema",
        )
        #   - dumper
        self.set(
            model,
            "record-dumper-class",
            lambda: f"{model.record_records_package}.dumper.{record_prefix}Dumper",
        )
        #   - search

        self.set(
            model,
            "record-search-options-class",
            lambda: f"{model.record_services_package}.search.{record_prefix}SearchOptions",
        )

        #   - facets
        self.set(
            model,
            "record-facets-class",
            lambda: f"{model.record_services_package}.facets.Test",
        )

        # alembic
        self.set(
            model,
            "record-schema-metadata-alembic",
            lambda: f"{extension_suffix}",
        )
        self.set(
            model,
            "record-schema-metadata-setup-cfg",
            lambda: f"{extension_suffix}",
        )

        self.set(
            model,
            "flask-commands-setup-cfg",
            lambda: f"{extension_suffix}",
        )

        self.set(model, "record-resource-blueprint-name", lambda: record_prefix)
        self.set(
            model,
            "create-blueprint-from-app",
            lambda: f"{model.package}.views.create_blueprint_from_app_{extension_suffix}",
        )
        model.setdefault("invenio-config-extra-code", "")
        model.setdefault("invenio-ext-extra-code", "")
        model.setdefault("invenio-proxies-extra-code", "")
        model.setdefault("invenio-record-extra-code", "")
        model.setdefault("invenio-record-dumper-extra-code", "")
        model.setdefault("invenio-record-facets-extra-code", "")
        model.setdefault("invenio-record-metadata-extra-code", "")
        model.setdefault("invenio-record-object-schema-extra-code", "")
        model.setdefault("invenio-record-permissions-extra-code", "")
        model.setdefault("invenio-record-resource-extra-code", "")
        model.setdefault("invenio-record-resource-config-extra-code", "")
        model.setdefault("invenio-record-schema-extra-code", "")
        model.setdefault("invenio-record-search-options-extra-code", "")
        model.setdefault("invenio-record-service-extra-code", "")
        model.setdefault("invenio-record-service-config-extra-code", "")
        model.setdefault("invenio-version-extra-code", "")
        model.setdefault("invenio-views-extra-code", "")

        if schema.model_field in schema.schema:
            current_model_field = schema.schema[schema.model_field]
            schema_class = model.record_schema_class
            schema_metadata_class = model.record_schema_metadata_class
            schema_class_base_classes = model.get(
                "record_schema_metadata_bases", []
            ) + [
                MARSHMALLOW_SCHEMA_BASE_CLASS  # alias will be recognized automatically
            ]

            if (
                "properties" in current_model_field
                and "metadata" in current_model_field.properties
            ):
                if current_model_field.properties.metadata.get("type") == "object":
                    deepmerge(
                        current_model_field.properties.metadata.setdefault(
                            "marshmallow", {}
                        ),
                        {
                            "schema-class": schema_metadata_class,
                            "base-classes": schema_class_base_classes,
                            "generate": True,
                        },
                    )

                    # ui-level of metadata
                    ui_schema_class = model.record_ui_schema_metadata_class
                    deepmerge(
                        current_model_field.properties.metadata.setdefault(
                            "ui", {}
                        ).setdefault("marshmallow", {}),
                        {
                            "schema-class": ui_schema_class,
                            "base-classes": [MARSHMALLOW_SCHEMA_BASE_CLASS],
                            "generate": True,
                        },
                    )

            if (
                "marshmallow" in schema.schema
                and "base-schema" in schema.schema["marshmallow"]
            ):
                schema_class_base_classes = model.get(
                    "record_schema_metadata_bases", []
                ) + [
                    schema.schema["marshmallow"][
                        "base_schema"
                    ]  # alias will be recognized automatically
                ]

            deepmerge(
                current_model_field.setdefault("marshmallow", {}),
                {
                    "schema-class": schema_class,
                    "base-classes": schema_class_base_classes,
                    "generate": True,
                },
            )
            # ui-level of current_model_field
            ui_schema_class = model.record_ui_schema_class
            deepmerge(
                current_model_field.setdefault("ui", {}).setdefault("marshmallow", {}),
                {
                    "schema-class": ui_schema_class,
                    "base-classes": [MARSHMALLOW_SCHEMA_BASE_CLASS],
                    "generate": True,
                },
            )
            current_model_field.setdefault("type", "object")

        model.setdefault("generate-record-pid-field", True)
        model.setdefault("record-dumper-extensions", [])

        # default import prefixes
        _prefixes = settings.python.setdefault("always-defined-import-prefixes", [])
        for _prefix in ["ma", "ma_fields", "ma_validate"]:
            if _prefix not in _prefixes:
                _prefixes.append(_prefix)

        model.setdefault("script-import-sample-data", "data/sample_data.yaml")
        self.set(model, "service-id", lambda: model.flask_extension_name)

        model.setdefault("permissions", {"presets": []})

        model.setdefault("pid-field-cls", "PIDField")
        model.setdefault("pid-field-provider", "RecordIdProviderV2")
        model.setdefault("pid-field-context", "PIDFieldContext")
        model.setdefault("pid-field-args", ["create=True"])
        model.setdefault(
            "pid-field-imports",
            [
                {
                    "import": "invenio_records_resources.records.systemfields.pid.PIDField"
                },
                {
                    "import": "invenio_records_resources.records.systemfields.pid.PIDFieldContext"
                },
                {"import": "invenio_pidstore.providers.recordid_v2.RecordIdProviderV2"},
            ],
        )



from oarepo_model_builder.utils.json_pathlib import JSONPaths, PathCondition


def test_condition():
    current = PathCondition(start={"a": {"b": 1}})
    assert current.a.b._subtree_list == [1]

    assert current.a.b == 1
    assert not (current.a.b == 2)
    assert current.a.b != 2


def test_condition_star():
    current = PathCondition(start={"a": {"b": 1, "c": 2}})
    assert current.a["*"]._subtree_list == [1, 2]

    assert current.a["*"] == 1
    assert current.a["*"] == 2
    assert current.a["*"] != 3


def test_condition_double_star():
    current = PathCondition(start={"a": {"b": 1, "c": {"d": 2}}})
    assert current.a["**"]._subtree_list == [1, {"d": 2}, 2]


def test_path_simple():
    p = JSONPaths()
    p.register(path="/a/b", value=1)
    assert list(p.match("/a/b")) == [1]
    assert list(p.match("/a")) == []


def test_path_locator():
    p = JSONPaths()
    p.register(path="/a/b", condition=lambda current: current.a == 1, value=1)
    assert list(p.match("/a/b", {"a": 1})) == [1]
    assert list(p.match("/a/b", {"a": 2})) == []


def test_path_multiple_locators():
    p = JSONPaths()
    p.register(path="/a/b", condition=lambda current: current.a == 1, value=1)
    p.register(path="/a/b", condition=lambda current: current.b == 1, value=2)
    assert list(p.match("/a/b", {"a": 1})) == [1]
    assert list(p.match("/a/b", {"b": 1})) == [2]
    assert list(p.match("/a/b", {"a": 1, "b": 1})) == [1, 2]
    assert list(p.match("/a/b", {"a": 2})) == []



from .json import JSONOutput


class JSONSchemaOutput(JSONOutput):
    TYPE = "jsonschema"

    def merge_jsonschema(self, jsonschema):
        self.stack.merge(jsonschema)

    def collect_required(self):
        top = self.stack.real_top
        # just a sanity check
        if not isinstance(top, dict):
            return
        if "properties" not in top:
            return
        required = []
        for prop_key, prop in top["properties"].items():
            if not isinstance(prop, dict):
                continue
            if not isinstance(prop.get("required"), list):
                if prop.pop("required", None):
                    required.append(prop_key)
        if required:
            required.sort()
            top["required"] = required

    def remove_required(self):
        top = self.stack.real_top
        # just a sanity check
        if not isinstance(top, dict):
            return
        if "properties" not in top:
            return
        required = []
        for prop_key, prop in top["properties"].items():
            if not isinstance(prop, dict):
                continue
            if not isinstance(prop.get("required"), list):
                if prop.pop("required", None):
                    required.append(prop_key)
        required.sort()
        return required



from oarepo_model_builder.builders import process
from oarepo_model_builder.utils.jinja import package_name

from ..datatypes import datatypes
from ..outputs.json_stack import JSONStack
from ..utils.deepmerge import deepmerge
from .invenio_base import InvenioBaseClassPythonBuilder

OAREPO_FACETS_PROPERTY = "facets"
OAREPO_SORTABLE_PROPERTY = "sortable"


class InvenioRecordSearchOptionsBuilder(InvenioBaseClassPythonBuilder):
    TYPE = "invenio_record_search"
    class_config = "record-search-options-class"
    template = None

    def begin(self, schema, settings):
        super().begin(schema, settings)
        self.template = "record-search-options"
        self.search_options_data = []
        self.sort_options_data = []
        self.search_options_stack = JSONStack()
        self.facets_definition = []
        self.settings = settings
        if "sortable" in schema.schema:
            self.process_top_sortable(schema.schema["sortable"])
        self.facet_switch = self.schema.model.get("searchable", True)
        self.facet_stack = []

    def finish(self, **extra_kwargs):
        super().finish(
            search_options_data=self.search_options_data,
            facets_definition=self.facets_definition,
            sort_definition=self.sort_options_data,
        )
        python_path = self.class_to_path(self.current_model.record_facets_class)
        self.process_template(
            python_path,
            "record-facets",
            current_package_name=package_name(self.current_model.record_facets_class),
            search_options_data=self.search_options_data,
            **extra_kwargs,
        )

    def process_top_sortable(self, data):
        keys = dir(data)
        for k in keys:
            fields = dir(data[k])
            fields_options = {}
            for field in fields:
                fields_options = deepmerge(fields_options, {field: data[k][field]})
            self.sort_options_data.append({k: fields_options})

    @process("**", condition=lambda current, stack: stack.schema_valid)
    def enter_model_element(self):
        schema_element_type = self.stack.top.schema_element_type
        data = self.stack.top.data
        recurse = True

        if recurse:
            self.build_children()

        if not self.search_options_stack:
            return

        if schema_element_type == "property":
            sort_definition = data.get(OAREPO_SORTABLE_PROPERTY, None)

            if sort_definition != None:
                self.sort_options_data.append(
                    self.process_sort_options(self.stack.path, sort_definition)
                )
        if self.stack.top.schema_element_type in ["property", "items"]:
            d_type = datatypes.get_datatype(
                self.stack.top.data,
                self.stack.top.key,
                self.current_model,
                self.schema,
                self.stack,
            )

            if d_type.get_facet(None, ""):
                stack_data = []
                for s in self.stack.stack:
                    type = s.schema_element_type
                    if type and type == "property":
                        stack_data.append(
                            datatypes.get_datatype(
                                s.data,
                                s.key,
                                self.current_model,
                                self.schema,
                                self.stack,
                            )
                        )
                    elif type == "items":
                        stack_data.append(
                            datatypes.get_datatype(
                                s.data,
                                None,
                                self.current_model,
                                self.schema,
                                self.stack,
                            )
                        )
                skip = False

                if self.facet_switch:
                    for s in stack_data:
                        if (
                            "facets" in s.definition
                            and "searchable" in s.definition["facets"]
                            and not s.definition["facets"]["searchable"]
                        ):
                            skip = True
                            break
                else:
                    skip = True
                    for s in stack_data:
                        if "facets" in s.definition:
                            if (
                                "searchable" in s.definition["facets"]
                                and s.definition["facets"]["searchable"]
                            ):
                                skip = False
                                break
                            elif (
                                "key" in s.definition["facets"]
                                or "field" in s.definition["facets"]
                            ):
                                skip = False
                                break

                if not skip:
                    # facet, path = datatypes.facet(stack_data)
                    facet_obj = datatypes.facet(stack_data)
                    for f in facet_obj:
                        facet = f['facet']
                        path = f['path']
                        self.search_options_data.append({path: facet})
                        search_ops_name = "facets." + path
                        self.facets_definition.append({path: search_ops_name})

    def process_name(self, path, type):
        path_array = (path.split("/"))[2:]
        name = path_array[0]
        if len(path_array) == 1:
            return name
        path_array.pop(0)
        last_path = ""
        for path in path_array:
            if last_path != "properties" and path == "items":
                continue
            last_path = path
            if path == "properties":
                continue
            if type == "name":
                name = name + "_" + path
            elif type == "field":
                name = name + "." + path

        return name

    def process_sort_options(self, path, definition):
        field = self.process_name(path=path, type="field")

        key = definition.get("key", "")
        if key == "":
            key = self.process_name(path, "name")
        order = definition.get("order", "asc")
        if order == "desc":
            field = "-" + field

        return {key: dict(fields=[field])}



from pkg_resources import parse_version

from oarepo_model_builder.builders import OutputBuilder
from oarepo_model_builder.outputs.cfg import CFGOutput
from oarepo_model_builder.utils.verbose import log


class SetupCfgBuilder(OutputBuilder):
    TYPE = "setup_cfg"

    def finish(self):
        super().finish()

        output: CFGOutput = self.builder.get_output("cfg", "setup.cfg")
        output.setdefault(
            "metadata", "name", self.current_model.package_base.replace("_", "-")
        )
        version = self.schema.get("version", "1.0.0dev1")
        output.setdefault("metadata", "version", version)
        if parse_version(output.get("metadata", "version").value) < parse_version(
            version
        ):
            output.set("metadata", "version", version)
        output.setdefault(
            "metadata",
            "description",
            f"A sample application for {self.current_model.package}",
        )
        output.setdefault("metadata", "authors", "")

        output.setdefault("options", "python", ">=3.9")

        output.add_dependency("invenio_access", ">=1.4.4")
        output.add_dependency("invenio_app", ">=1.3.4")
        output.add_dependency("invenio_db", ">=1.0.14")
        output.add_dependency("invenio_pidstore", ">=1.2.3")
        output.add_dependency("invenio_records", ">=2.0.0")
        output.add_dependency("invenio-records-rest", ">=2.1.0")
        output.add_dependency("invenio_records_permissions", ">=0.13.0")
        output.add_dependency("invenio_records_resources", ">=0.21.4")
        output.add_dependency("invenio-search", ">=2.1.0")
        output.add_dependency("tqdm", ">=4.64.1")
        output.add_dependency("oarepo-runtime", ">=1.0.0")

        output.setdefault("options", "packages", "find:")

        output.setdefault(
            "options.package_data", "*", "*.json, *.rst, *.md, *.json5, *.jinja2"
        )

        if "runtime-dependencies" in self.schema:
            for dep, value in self.schema.runtime_dependencies.items():
                output.add_dependency(dep, ">=" + value)

        if "dev-dependencies" in self.schema:
            for dep, value in self.schema.dev_dependencies.items():
                output.add_dependency(
                    dep, ">=" + value, group="options.extras_require", section="devs"
                )

        if output.created:
            log(
                log.INFO,
                f"""To install the data model, run
    poetry install -E {self.current_model.package}            
            """,
            )



from .invenio_base import InvenioBaseClassPythonBuilder


class InvenioViewsBuilder(InvenioBaseClassPythonBuilder):
    TYPE = "invenio_views"
    class_config = "create_blueprint_from_app"
    template = "views"



from oarepo_model_builder.model_preprocessors import ModelPreprocessor
from oarepo_model_builder.utils.deepmerge import deepmerge


class OpensearchModelPreprocessor(ModelPreprocessor):
    TYPE = "opensearch"

    def transform(self, schema, settings):
        deepmerge(
            settings,
            {"opensearch": {"version": "os-v2"}},
        )



from typing import List

from ..utils.facet_helpers import facet_definition, facet_name
from .datatypes import DataType, Import


class BaseDateDataType(DataType):
    marshmallow_field = "ma_fields.String"

    def get_facet(self, stack, parent_path):
        key, field = facet_definition(self)
        path = parent_path
        if len(parent_path) > 0 and self.key:
            path = parent_path + "." + self.key
        elif self.key:
            path = self.key
        if field:
            return [{"facet":field, "path":  facet_name(path)}]
        else:
            return [{"facet" : f'TermsFacet(field="{path}")', "path" : facet_name(path)}]


class DateDataType(BaseDateDataType):
    schema_type = "string"
    mapping_type = "date"
    ui_marshmallow_field = "l10n.LocalizedDate"
    model_type = "date"

    def mapping(self, **extras):
        return super().mapping(format="strict_date", **extras)

    def json_schema(self, **extras):
        return super().json_schema(format="date", **extras)

    def marshmallow_validators(self) -> List[str]:
        return super().marshmallow_validators() + [
            "validate_date('%Y-%m-%d')",
        ]

    def imports(self, *extra) -> List[Import]:
        return super().imports(
            Import(import_path="oarepo_runtime.validation.validate_date", alias=None),
            Import(import_path="oarepo_runtime.ui.marshmallow", alias="l10n"),
        )


class TimeDataType(BaseDateDataType):
    schema_type = "string"
    mapping_type = "date"
    model_type = "time"
    ui_marshmallow_field = "l10n.LocalizedTime"

    def mapping(self, **extras):
        return super().mapping(format="strict_time||strict_time_no_millis")

    def json_schema(self, **extras):
        return super().json_schema(format="time", **extras)

    def marshmallow_validators(self) -> List[str]:
        return super().marshmallow_validators() + [
            "validate_date('%H:%M:%S')",
        ]

    def imports(self, *extra) -> List[Import]:
        return super().imports(
            Import(import_path="oarepo_runtime.validation.validate_date", alias=None),
            Import(import_path="oarepo_runtime.ui.marshmallow", alias="l10n"),
        )


class DateTimeDataType(BaseDateDataType):
    schema_type = "string"
    mapping_type = "date"
    marshmallow_field = "ma_fields.String"
    model_type = "datetime"
    ui_marshmallow_field = "l10n.LocalizedDateTime"

    def mapping(self, **extras):
        return super().mapping(format="strict_date_time||strict_date_time_no_millis")

    def json_schema(self, **extras):
        return super().json_schema(format="date-time", **extras)

    def marshmallow_validators(self) -> List[str]:
        return super().marshmallow_validators() + [
            "validate_datetime",
        ]

    def imports(self, *extra) -> List[Import]:
        return super().imports(
            Import(
                import_path="oarepo_runtime.validation.validate_datetime", alias=None
            ),
            Import(import_path="oarepo_runtime.ui.marshmallow", alias="l10n"),
        )


class EDTFDataType(BaseDateDataType):
    schema_type = "string"
    mapping_type = "date"
    model_type = "edtf"
    ui_marshmallow_field = "l10n.LocalizedEDTF"

    def mapping(self, **extras):
        return super().mapping(
            format="strict_date_time||strict_date_time_no_millis||strict_date||yyyy-MM||yyyy"
        )

    def marshmallow_validators(self):
        return super().marshmallow_validators() + [
            "mu_fields_edtf.EDTFValidator(types=(EDTFDate,))"
        ]

    def imports(self, *_extra) -> List[Import]:
        return super().imports(
            Import(import_path="oarepo_runtime.ui.marshmallow", alias="l10n"),
            Import(import_path="edtf.Date", alias="EDTFDate"),
            Import(
                import_path="marshmallow_utils.fields.edtfdatestring",
                alias="mu_fields_edtf",
            ),
        )


class EDTFIntervalType(BaseDateDataType):
    schema_type = "string"
    mapping_type = "date_range"
    model_type = "edtf-interval"
    ui_marshmallow_field = "l10n.LocalizedEDTFInterval"

    def mapping(self, **extras):
        return super().mapping(
            format="strict_date_time||strict_date_time_no_millis||strict_date||yyyy-MM||yyyy"
        )

    def marshmallow_validators(self):
        return super().marshmallow_validators() + [
            "mu_fields_edtf.EDTFValidator(types=(EDTFInterval,))"
        ]

    def imports(self, *extra):
        return super().imports(
            Import(import_path="edtf.Interval", alias="EDTFInterval"),
            Import(import_path="oarepo_runtime.ui.marshmallow", alias="l10n"),
            Import(
                import_path="marshmallow_utils.fields.edtfdatestring",
                alias="mu_fields_edtf",
            ),
            *extra,
        )



import os
import re

import pytest

from oarepo_model_builder.builder import ModelBuilder
from oarepo_model_builder.fs import InMemoryFileSystem
from oarepo_model_builder.invenio.invenio_record_schema import (
    InvenioRecordSchemaBuilder,
)
from oarepo_model_builder.model_preprocessors.datatype_default import (
    DatatypeDefaultModelPreprocessor,
)
from oarepo_model_builder.model_preprocessors.default_values import (
    DefaultValuesModelPreprocessor,
)
from oarepo_model_builder.model_preprocessors.invenio import InvenioModelPreprocessor
from oarepo_model_builder.model_preprocessors.opensearch import (
    OpensearchModelPreprocessor,
)
from oarepo_model_builder.outputs.python import PythonOutput
from oarepo_model_builder.property_preprocessors.datatype_preprocessor import (
    DataTypePreprocessor,
)
from oarepo_model_builder.schema import ModelSchema

OAREPO_MARSHMALLOW = "marshmallow"
B_SCHEMA = 'classB(ma.Schema):"""Bschema."""b=ma_fields.String()'


def get_test_schema(**props):
    return ModelSchema(
        "",
        {
            "settings": {
                "python": {"use-isort": False, "use-black": False},
            },
            "model": {"package": "test", "properties": props},
        },
    )


@pytest.fixture
def fulltext_builder():
    return ModelBuilder(
        output_builders=[InvenioRecordSchemaBuilder],
        outputs=[PythonOutput],
        model_preprocessors=[
            DefaultValuesModelPreprocessor,
            OpensearchModelPreprocessor,
            InvenioModelPreprocessor,
            DatatypeDefaultModelPreprocessor,
        ],
        property_preprocessors=[DataTypePreprocessor],
    )


def _test(fulltext_builder, string_type):
    schema = get_test_schema(a={"type": string_type})
    fulltext_builder.filesystem = InMemoryFileSystem()
    fulltext_builder.build(schema, output_dir="")

    with fulltext_builder.filesystem.open(
        os.path.join("test", "services", "records", "schema.py")
    ) as f:
        data = f.read()
    assert "a = ma_fields.String()" in data


def test_fulltext(fulltext_builder):
    _test(fulltext_builder, "fulltext")


def test_keyword(fulltext_builder):
    _test(fulltext_builder, "keyword")


def test_fulltext_keyword(fulltext_builder):
    _test(fulltext_builder, "fulltext+keyword")


def test_simple_array(fulltext_builder):
    schema = get_test_schema(a={"type": "array", "items": {"type": "keyword"}})
    fulltext_builder.filesystem = InMemoryFileSystem()
    fulltext_builder.build(schema, output_dir="")

    with fulltext_builder.filesystem.open(
        os.path.join("test", "services", "records", "schema.py")
    ) as f:
        data = f.read()
    assert "a = ma_fields.List(ma_fields.String())" in data


def test_array_of_objects(fulltext_builder):
    schema = get_test_schema(
        a={
            "type": "array",
            "items": {"type": "object", "properties": {"b": {"type": "integer"}}},
        }
    )
    fulltext_builder.filesystem = InMemoryFileSystem()
    fulltext_builder.build(schema, output_dir="")

    with fulltext_builder.filesystem.open(
        os.path.join("test", "services", "records", "schema.py")
    ) as f:
        data = f.read()
    assert (
        'class AItemSchema(ma.Schema):\n    """AItemSchema schema."""\n    b = ma_fields.Integer()'
        in data
    )
    assert "a = ma_fields.List(ma_fields.Nested(lambda: AItemSchema()))" in data


def test_generate_nested_schema_same_file(fulltext_builder):
    schema = get_test_schema(
        a={
            "type": "object",
            OAREPO_MARSHMALLOW: {"schema-class": "B", "generate": True},
            "properties": {
                "b": {
                    "type": "keyword",
                }
            },
        }
    )
    fulltext_builder.filesystem = InMemoryFileSystem()
    fulltext_builder.build(schema, output_dir="")

    with fulltext_builder.filesystem.open(
        os.path.join("test", "services", "records", "schema.py")
    ) as f:
        data = f.read()
    assert (
        re.sub(
            r"\s",
            "",
            """class B(ma.Schema):
        \"""B schema.\"""
        
        b = ma_fields.String()""",
        )
        in re.sub(r"\s", "", data)
    )
    assert (
        re.sub(
            r"\s",
            "",
            """class TestSchema(ma.Schema):
        \"""TestSchema schema.\"""
        
        a = ma_fields.Nested(lambda: B())""",
        )
        in re.sub(r"\s", "", data)
    )


def test_generate_nested_schema_different_file(fulltext_builder):
    schema = get_test_schema(
        a={
            "type": "object",
            OAREPO_MARSHMALLOW: {
                "schema-class": "test.services.schema2.B",
                "generate": True,
            },
            "properties": {
                "b": {
                    "type": "keyword",
                }
            },
        }
    )
    fulltext_builder.filesystem = InMemoryFileSystem()
    fulltext_builder.build(schema, output_dir="")

    with fulltext_builder.filesystem.open(
        os.path.join("test", "services", "records", "schema.py")
    ) as f:
        data = f.read()

    assert (
        re.sub(
            r"\s",
            "",
            """class TestSchema(ma.Schema):
        \"""TestSchema schema.\"""
    
        a = ma_fields.Nested(lambda: B())""",
        )
        in re.sub(r"\s", "", data)
    )

    assert "from test.services.schema2 import B" in data

    with fulltext_builder.filesystem.open(
        os.path.join("test", "services", "schema2.py")
    ) as f:
        data = f.read()
    assert B_SCHEMA in re.sub(r"\s", "", data)


def test_use_nested_schema_same_file(fulltext_builder):
    schema = get_test_schema(
        a={
            "type": "object",
            OAREPO_MARSHMALLOW: {"schema-class": "B", "generate": False},
            "properties": {
                "b": {
                    "type": "keyword",
                }
            },
        }
    )
    fulltext_builder.filesystem = InMemoryFileSystem()
    fulltext_builder.build(schema, output_dir="")

    with fulltext_builder.filesystem.open(
        os.path.join("test", "services", "records", "schema.py")
    ) as f:
        data = f.read()
        print(data)
    assert (
        re.sub(
            r"\s",
            "",
            """class TestSchema(ma.Schema):
        \"""TestSchema schema.\"""
        
        a = ma_fields.Nested(lambda: B())""",
        )
        in re.sub(r"\s", "", data)
    )
    assert "classB(ma.Schema)" not in re.sub(r"\s", "", data)


def test_use_nested_schema_different_file(fulltext_builder):
    schema = get_test_schema(
        a={
            "type": "object",
            OAREPO_MARSHMALLOW: {"schema-class": "c.B", "generate": False},
            "properties": {
                "b": {
                    "type": "keyword",
                }
            },
        }
    )
    fulltext_builder.filesystem = InMemoryFileSystem()
    fulltext_builder.build(schema, output_dir="")

    with fulltext_builder.filesystem.open(
        os.path.join("test", "services", "records", "schema.py")
    ) as f:
        data = f.read()
    assert re.sub(r"\s", "", "from c import B") in re.sub(r"\s", "", data)
    assert (
        'classTestSchema(ma.Schema):"""TestSchemaschema."""a=ma_fields.Nested(lambda:B())'
        in re.sub(r"\s", "", data)
    )


def test_generate_nested_schema_array(fulltext_builder):
    schema = get_test_schema(
        a={
            "type": "array",
            "items": {
                "type": "object",
                OAREPO_MARSHMALLOW: {"schema-class": "B", "generate": True},
                "properties": {
                    "b": {
                        "type": "keyword",
                    }
                },
            },
        }
    )
    fulltext_builder.filesystem = InMemoryFileSystem()
    fulltext_builder.build(schema, output_dir="")

    with fulltext_builder.filesystem.open(
        os.path.join("test", "services", "records", "schema.py")
    ) as f:
        data = f.read()
    assert B_SCHEMA in re.sub(r"\s", "", data)
    assert (
        'classTestSchema(ma.Schema):"""TestSchemaschema."""a=ma_fields.List(ma_fields.Nested(lambda:B()))'
        in re.sub(r"\s", "", data)
    )


def test_extend_existing(fulltext_builder):
    schema = get_test_schema(a={"type": "keyword"}, b={"type": "keyword"})
    fulltext_builder.filesystem = InMemoryFileSystem()

    with fulltext_builder.filesystem.open(
        os.path.join("test", "services", "records", "schema.py"), "w"
    ) as f:
        f.write(
            '''
from invenio_records_resources.services.records.schema import BaseRecordSchema as InvenioBaseRecordSchema
import marshmallow as ma
import marshmallow.fields as ma_fields
import marshmallow.validate as ma_valid
class TestSchema(ma.Schema):
    """TestSchema schema."""
    a = ma_fields.String()'''
        )

    fulltext_builder.build(schema, output_dir="")
    data = fulltext_builder.filesystem.read(
        os.path.join("test", "services", "records", "schema.py")
    )
    assert "b = ma_fields.String()" in data


def test_generate_nested_schema_relative_same_package(fulltext_builder):
    schema = get_test_schema(
        a={
            "type": "object",
            OAREPO_MARSHMALLOW: {
                "schema-class": ".schema2.B",
                "generate": True,
            },
            "properties": {
                "b": {
                    "type": "keyword",
                }
            },
        }
    )
    fulltext_builder.filesystem = InMemoryFileSystem()
    fulltext_builder.build(schema, output_dir="")

    with fulltext_builder.filesystem.open(
        os.path.join("test", "services", "records", "schema.py")
    ) as f:
        data = f.read()

    assert (
        re.sub(
            r"\s",
            "",
            """class TestSchema(ma.Schema):
        \"""TestSchema schema.\"""
    
        a = ma_fields.Nested(lambda:B())""",
        )
        in re.sub(r"\s", "", data)
    )

    assert "from test.services.records.schema2 import B" in data

    with fulltext_builder.filesystem.open(
        os.path.join("test", "services", "records", "schema2.py")
    ) as f:
        data = f.read()
    assert B_SCHEMA in re.sub(r"\s", "", data)


def test_generate_nested_schema_relative_same_file(fulltext_builder):
    schema = get_test_schema(
        a={
            "type": "object",
            OAREPO_MARSHMALLOW: {
                "schema-class": ".B",
                "generate": True,
            },
            "properties": {
                "b": {
                    "type": "keyword",
                }
            },
        }
    )
    fulltext_builder.filesystem = InMemoryFileSystem()
    fulltext_builder.build(schema, output_dir="")

    with fulltext_builder.filesystem.open(
        os.path.join("test", "services", "records", "schema.py")
    ) as f:
        data = f.read()

    assert (
        re.sub(
            r"\s",
            "",
            """class TestSchema(ma.Schema):
        \"""TestSchema schema.\"""

        a = ma_fields.Nested(lambda:B())""",
        )
        in re.sub(r"\s", "", data)
    )


def test_generate_nested_schema_relative_upper(fulltext_builder):
    schema = get_test_schema(
        a={
            "type": "object",
            OAREPO_MARSHMALLOW: {
                "schema-class": "..schema2.B",
                "generate": True,
            },
            "properties": {
                "b": {
                    "type": "keyword",
                }
            },
        }
    )
    fulltext_builder.filesystem = InMemoryFileSystem()
    fulltext_builder.build(schema, output_dir="")

    with fulltext_builder.filesystem.open(
        os.path.join("test", "services", "records", "schema.py")
    ) as f:
        data = f.read()

    assert (
        re.sub(
            r"\s",
            "",
            """class TestSchema(ma.Schema):
        \"""TestSchema schema.\"""

        a = ma_fields.Nested(lambda: B())""",
        )
        in re.sub(r"\s", "", data)
    )

    assert "from test.services.schema2 import B" in data

    with fulltext_builder.filesystem.open(
        os.path.join("test", "services", "schema2.py")
    ) as f:
        data = f.read()
    assert B_SCHEMA in re.sub(r"\s", "", data)



import os

import json5
import pytest

from oarepo_model_builder.builder import ModelBuilder
from oarepo_model_builder.builders.jsonschema import JSONSchemaBuilder
from oarepo_model_builder.builders.mapping import MappingBuilder
from oarepo_model_builder.fs import InMemoryFileSystem
from oarepo_model_builder.model_preprocessors.default_values import (
    DefaultValuesModelPreprocessor,
)
from oarepo_model_builder.model_preprocessors.opensearch import (
    OpensearchModelPreprocessor,
)
from oarepo_model_builder.outputs.jsonschema import JSONSchemaOutput
from oarepo_model_builder.outputs.mapping import MappingOutput
from oarepo_model_builder.outputs.python import PythonOutput
from oarepo_model_builder.property_preprocessors.datatype_preprocessor import (
    DataTypePreprocessor,
)
from oarepo_model_builder.schema import ModelSchema


def get_model_schema(field_type):
    return ModelSchema(
        "",
        {
            "settings": {
                "python": {"use-isort": False, "use-black": False},
            },
            "model": {"package": "test", "properties": {"a": {"type": field_type}}},
        },
    )


@pytest.fixture
def fulltext_builder():
    return ModelBuilder(
        output_builders=[JSONSchemaBuilder, MappingBuilder],
        outputs=[JSONSchemaOutput, MappingOutput, PythonOutput],
        model_preprocessors=[
            DefaultValuesModelPreprocessor,
            OpensearchModelPreprocessor,
        ],
        property_preprocessors=[DataTypePreprocessor],
    )


def test_fulltext(fulltext_builder):
    schema = get_model_schema("fulltext")
    fulltext_builder.filesystem = InMemoryFileSystem()
    fulltext_builder.build(schema, output_dir="")

    data = load_generated_jsonschema(fulltext_builder)

    assert data == {"type": "object", "properties": {"a": {"type": "string"}}}

    data = load_generated_mapping(fulltext_builder)

    assert data == {
        "mappings": {
            "properties": {
                "a": {"type": "text"},
            }
        }
    }


def test_keyword(fulltext_builder):
    schema = get_model_schema("keyword")
    fulltext_builder.filesystem = InMemoryFileSystem()
    fulltext_builder.build(schema, output_dir="")

    data = load_generated_jsonschema(fulltext_builder)

    assert data == {"type": "object", "properties": {"a": {"type": "string"}}}

    data = load_generated_mapping(fulltext_builder)

    assert data == {
        "mappings": {
            "properties": {
                "a": {"type": "keyword"},
            }
        }
    }


def test_fulltext_keyword(fulltext_builder):
    schema = get_model_schema("fulltext+keyword")
    fulltext_builder.filesystem = InMemoryFileSystem()
    fulltext_builder.build(schema, output_dir="")

    data = load_generated_jsonschema(fulltext_builder)

    assert data == {"type": "object", "properties": {"a": {"type": "string"}}}

    data = load_generated_mapping(fulltext_builder)

    assert data == {
        "mappings": {
            "properties": {
                "a": {"fields": {"keyword": {"type": "keyword"}}, "type": "text"},
            }
        }
    }


def load_generated_mapping(fulltext_builder):
    return json5.load(
        fulltext_builder.filesystem.open(
            os.path.join(
                "test", "records", "mappings", "os-v2", "test", "test-1.0.0.json"
            )
        )
    )


def load_generated_jsonschema(fulltext_builder):
    return json5.load(
        fulltext_builder.filesystem.open(
            os.path.join("test", "records", "jsonschemas", "test-1.0.0.json")
        )
    )



from pathlib import Path

from ..utils.verbose import log
from . import process
from .json_base import JSONBaseBuilder
from .utils import ensure_parent_modules


class MappingBuilder(JSONBaseBuilder):
    TYPE = "mapping"
    output_file_type = "mapping"
    output_file_name = "mapping-file"
    parent_module_root_name = "mappings"

    @process("**", condition=lambda current, stack: stack.schema_valid)
    def enter_model_element(self):
        # ignore schema leaves different than "type" - for example, minLength, ...
        # that should not be present in mapping
        element_type = self.stack.top.schema_element_type

        if element_type == "items":
            # do not output "items" container
            self.build_children()
            self.set_searchable()
            self.merge_mapping(self.stack.top.data)

        elif element_type in ("properties", "property"):
            self.index_enabled_stack.append(
                self.stack.top.data.get("facets", {}).get(
                    "searchable", self.index_enabled_stack[-1]
                )
            )
            self.children_indexed.append(False)

            self.model_element_enter()

            # process children
            self.build_children()
            if element_type == "property":
                self.set_searchable()

            self.merge_mapping(self.stack.top.data)

            self.model_element_leave()

            self.index_enabled_stack.pop()
            self.children_indexed.pop()

        elif element_type == "type" and self.stack.top.data != "array":
            # do not output "type=array"
            self.model_element_enter()
            self.model_element_leave()

    def set_searchable(self):
        if self.stack.top.data.type != "array":
            if self.stack.top.data.type in ("object", "nested"):
                searchable = self.children_indexed[-1]
                if not searchable:
                    self.stack.top.data.setdefault("mapping", {})["enabled"] = False
            else:
                searchable = self.index_enabled_stack[-1] or self.children_indexed[-1]
                if not searchable:
                    self.stack.top.data.setdefault("mapping", {})["index"] = False
                else:
                    for idx in range(0, len(self.children_indexed)):
                        self.children_indexed[idx] = True

    def merge_mapping(self, data):
        if isinstance(data, dict) and "mapping" in data:
            mapping = self.call_components(
                "before_merge_mapping", data["mapping"], stack=self.stack
            )
            if not mapping.get("enabled", True):
                self.output.replace_mapping(mapping)
            else:
                self.output.merge_mapping(mapping)

    def on_enter_model(self, output_name):
        self.index_enabled_stack = [self.current_model.get("searchable", True)]
        self.children_indexed = [False]
        ensure_parent_modules(
            self.builder, Path(output_name), ends_at=self.parent_module_root_name
        )
        if "mapping" in self.current_model:
            if (
                "opensearch" not in self.settings
                or "version" not in self.settings.opensearch
            ):
                raise ValueError(
                    "Please define settings.opensearch.version (for example, to os-v2)"
                )
            self.output.merge(
                self.current_model.mapping[self.settings.opensearch.version]
            )
        self.output.enter("mappings", {})

    def finish(self):
        super().finish()

        log(
            log.INFO,
            f"""
    invenio index init --force
            """,
        )



from .invenio_base import InvenioBaseClassPythonBuilder


class InvenioRecordResourceConfigBuilder(InvenioBaseClassPythonBuilder):
    TYPE = "invenio_record_resource_config"
    class_config = "record-resource-config-class"
    template = "record-resource-config"



from marshmallow import fields

from .utils import ExtendablePartSchema


class PropertyFacets(ExtendablePartSchema):
    key = fields.String(required=False)
    field = fields.String(required=False)
    searchable = fields.Boolean(required=False)



from .invenio_base import InvenioBaseClassPythonBuilder


class InvenioRecordBuilder(InvenioBaseClassPythonBuilder):
    TYPE = "invenio_record"
    class_config = "record-class"
    template = "record"



from .python_name import convert_name_to_python


def facet_definition(obj):
    key = None
    field = None
    if "facets" in obj.definition:
        key = obj.definition["facets"].get("key", obj.key)
        field = obj.definition["facets"].get("field", None)
    if not key:
        key = obj.key
    return key, field


def facet_name(path):
    path.replace(".", "_")
    path = convert_name_to_python(path)
    return path
